# -*- coding: utf-8 -*-
# Standard Libraries
from concurrent.futures import ThreadPoolExecutor
from datetime import date
from os import getcwd, cpu_count
from os.path import isdir, join
from typing import List, Dict, Union, Optional

# PyPi Packages
from geopandas import GeoDataFrame
from pandas import concat
from shapely.geometry import Polygon

# Local Modules
from .cbers4a import (
    Search,
    Download,
    SearchItem,
    ItemCollection,
    Item,
    Collections,
)


class Cbers4aAPI:
    """
    The CBERS4A API class. Query, download or transform data from CBERS-4A STAC API.

    Args:
        email: Sign-in e-mail used at https://www.dgi.inpe.br/catalogo/explore
    """

    def __init__(self, email: Optional[str] = None):
        self._email = email

    @property
    def email(self):
        return self._email

    @email.setter
    def email(self, email: str):
        self._email = email

    @staticmethod
    def query(
        location: Union[list[float], Polygon, tuple],
        initial_date: date,
        end_date: date,
        cloud: int,
        limit: int,
        collections: Union[list[str], list[Collections]],
    ) -> dict:
        """
        Query Images from INPE's catalog

        Args:
            location: Bounding box, Polygon shape or Path and Row Tuple
            initial_date: Images from this date
            end_date: Images to this date
            cloud: Percentage of cloud coverage
            limit: Limit of returned images
            collections: Collection's name(s)
        Notes:
            Location:
                - Bounding box: `location=[-0.5, 1.0, 0.5, -0.5]`
                - Polygon shape from shapely
                - Path and row respectively: `location=(225,75)`
            Initial and end dates:
                - Use date object from datetime library: `datetime.date(2021,05,25)`
                - Caution to end date always be bigger than initial date
                - e.g. `initial_date=datetime.date(2021,08,15)`
            Cloud coverage:
                - Search for scenes less or equal (<=) cloud coverage percentage.
                - e.g. `cloud=70`
                - Must be between 0 and 100.
            Limit:
                - Values greater than 0.
                - e.g. `limit=500`.
            Collections:
                - Use the collections enum via `from cbers4asat import Collections as col`
                - Available collections from INPE catalog <https://www.dgi.inpe.br/catalogo/explore>
                - Always an array of collections name.
                - e.g. `[col.AMAZONIA1_WFI_L2_DN]` or `[col.AMAZONIA1_WFI_L2_DN, col.CBERS4A_WPM_L4_DN]` etc.

        Returns:
            dict: Dict with GeoJSON-like format
        Raises:
            Exception: If any input is invalid.
        """
        search = Search()

        if isinstance(location, list):
            search.bbox(location)
        elif isinstance(location, Polygon):
            search.bbox(list(location.bounds))
        elif isinstance(location, tuple):
            search.path_row(*location)
        else:
            raise Exception(
                "Provide a Bouding box, Polygon shape or path and row tuple"
            )

        search.date_interval(initial_date, end_date)
        search.cloud_cover(cloud)
        search.limit(limit)
        search.collections(collections)

        return search()

    @staticmethod
    def query_by_id(
        scene_id: Union[List[str], str], collection: Union[str, Collections]
    ):
        """
        Search a product by id

        Args:
            scene_id: One or more scene's id
            collection: Collection's name
        Returns:
            dict: Dict with GeoJSON-like format
        """
        search = SearchItem()
        search.ids(
            scene_id if isinstance(scene_id, list) else list([scene_id]),
            collection=collection,
        )
        return search()

    def __download(
        self,
        products: Dict,
        bands: List[str],
        threads: int = cpu_count(),
        outdir: str = getcwd(),
        with_folder: bool = False,
        with_metadata: bool = False,
    ):
        try:
            products = ItemCollection(**products)
        except TypeError:
            raise Exception(
                "Check your product structure. It must be a GeoJSON like dictionary."
            )

        products.get_features_assets()

        tasks = list()
        root = outdir
        for product in products:
            if with_folder:
                outdir = join(root, product.id)

            for band in bands:
                tasks.append(
                    (Download().download, product.band_url(band), self.email, outdir)
                )
                if with_metadata:
                    url_xml = product.band_url(band).replace('.tif', '.xml')
                    tasks.append(
                        (Download().download, url_xml, self.email, outdir)
                    )

        with ThreadPoolExecutor(max_workers=threads) as t_pool:
            for task in tasks:
                f = t_pool.submit(*task)
                if f.exception():
                    raise f.exception()

    def __download_gdf(
        self,
        products: GeoDataFrame,
        bands: List[str],
        threads: int = cpu_count(),
        outdir: str = getcwd(),
        with_folder: bool = False,
        with_metadata: bool = False,
    ):
        features = list()
        for index, row in products.iterrows():
            features.append(Item.from_search(row.id, row.collection))

        # GeoDataFrame is not needed anymore
        products = ItemCollection(type="FeatureCollection", features=features)

        tasks = list()
        root = outdir
        for product in products:
            if with_folder:
                outdir = join(root, product.id)
            for band in bands:
                tasks.append(
                    (Download().download, product.band_url(band), self.email, outdir)
                )
                if with_metadata:
                    url_xml = product.band_url(band).replace('.tif', '.xml')
                    tasks.append(
                        (Download().download, url_xml, self.email, outdir)
                    )


        with ThreadPoolExecutor(max_workers=threads) as t_pool:
            for task in tasks:
                f = t_pool.submit(*task)
                if f.exception():
                    raise f.exception()

    def download(
        self,
        products: Union[dict, GeoDataFrame],
        bands: list[str],
        threads: int = cpu_count(),
        outdir: str = getcwd(),
        with_folder: bool = False,
        with_metadata: bool = False,
    ):
        """
        Download bands from all given scenes

        Args:
            products: Data returned from API
            bands: List of band color's name. "red", "green", "blue", "nir", "pan"
            threads: Max of cpu threads
            outdir: Output path
            with_folder: Group scene bands in a sub folder
            with_metadata: Download band's metadata (XML)
        Examples:
            - download(my_query_result, ['red', 'green'], 3, './downlaods', true)
            - download(my_query_result, ['red'], outdir='./downloads', with_folder=true)
            - download(my_query_result, ['blue'], with_metadata=True)
        Returns:
            GeoTIFF files
        """
        if not len(bands):
            raise TypeError("Choose bands to download.")
        elif not isdir(outdir):
            raise NotADirectoryError("Choose a valid output directory.")
        elif not self.email:
            raise Exception("Credentials not provided!")

        if isinstance(products, dict):
            if not products:  # Check if dictionary is empty
                raise Exception("No product to download.")
            return self.__download(products, bands, threads, outdir, with_folder, with_metadata)
        elif isinstance(products, GeoDataFrame):
            if products.empty:  # Check if data frame is empty
                raise Exception("No product to download.")
            return self.__download_gdf(products, bands, threads, outdir, with_folder, with_metadata)
        else:
            raise Exception("Bad Arguments.")

    @staticmethod
    def to_geodataframe(products: dict) -> GeoDataFrame:
        """
        Transform products list to a GeoDataFrame

        Args:
            products: GeoJSON-like dictionary returned from API
        Returns:
            GeoDataFrame of products.
        """
        if not products or not isinstance(products, dict):
            raise Exception("Provide a valid product structure.")

        item_collection = ItemCollection(**products)
        processed = list()
        for item in item_collection:
            extras = {
                "properties": {
                    **item.properties.asdict(),
                    "id": item.id,
                    "bbox": item.bbox,
                    "collection": item.collection,
                    "thumbnail": item.assets.thumbnail.href,
                }
            }
            gdf = GeoDataFrame.from_features(
                {
                    "type": "FeatureCollection",
                    "features": [item.asdict() | extras],
                },
                crs="EPSG:4326",
            )
            processed.append(gdf)

        return concat(processed, ignore_index=True).set_index("id", drop=False)



from os import getcwd
from os.path import basename, join
from typing import Literal
from requests import Session, HTTPError


def grid_download(
    satellite: Literal["cbers4a", "amazonia1"] = "cbers4a",
    sensor: Literal["mux", "wfi"] = "mux",
    outdir: str = getcwd(),
) -> None:
    """
    Download path and row grid from CBERS-04A and AMAZONIA1

    Args:
        satellite: "cbers4a" or "amazonia1"
        sensor: "mux" or "wfi"
        outdir: Output path
    Examples:
        - grid_download("cbers4a", "mux")
        - grid_download(satellite="amazonia1", "wfi", outdir="./downloads")
    Returns:
        .kmz file
    """
    grids = {
        "cbers4a": {
            "mux": "http://www.obt.inpe.br/OBT/assuntos/catalogo-cbers-amz-1/grid_cbers4a_mux.kmz",
            "wfi": "http://www.obt.inpe.br/OBT/assuntos/catalogo-cbers-amz-1/grid_cbers4a_wfi.kmz",
        },
        "amazonia1": {
            "wfi": "http://www.obt.inpe.br/OBT/assuntos/catalogo-cbers-amz-1/grid_amazonia1_wfi_sa.kmz"
        },
    }

    get_satellite = grids.get(satellite.lower(), None)

    if get_satellite:
        get_sensor = get_satellite.get(sensor.lower(), None)

        if get_sensor:
            with Session() as session:
                try:
                    req = session.get(url=get_sensor, stream=True, allow_redirects=True)
                    req.raise_for_status()

                    filename = basename(get_sensor)
                    with open(join(outdir, filename), "wb") as f:
                        for chunk in req.iter_content(chunk_size=1024):
                            if chunk:
                                f.write(chunk)

                except HTTPError as err:
                    raise Exception(f"Download unavailable. Error: {err}")
        else:
            raise ValueError("Sensors available: mux and wfi")
    else:
        raise ValueError("Satellites available: cbers4a and amazonia1")



# -*- coding: utf-8 -*-
"""
This dataclasses inside this file will generate something like this when requesting data
to STAC API in Search class:

{
  "providers":[
    {
      "name": "LGI-CDSR",
      "collections": [
        {
          "name": "CBERS4A_WPM_L4_DN"
        },
      ],
      "method": "POST",
      "query": {"cloud_cover": {"lte": 100}}
    }
  ],
  "datetime": "2024-12-18T00:00:00/2025-01-18T23:59:00",
  "limit": 100,
  "bbox": [
    -63.93905639648438,
    -9.00445156167208,
    -63.44879150390626,
    -8.733077421211563
  ],
  "fromCatalog": "yes"
}
"""

from dataclasses import dataclass, field
from typing import List, Optional, Union
from .collections import Collections
from .utils.dataclass import SerializationCapabilities


@dataclass
class Collection:
    """
    Represents the collection inside collection list (providers.collections)
    """

    name: Union[str, Collections] = None


@dataclass
class Query:
    """
    Represents the values to query inside the LGI-CDSR provider. (providers.query)
    """

    cloud_cover: dict[str, int] = field(default_factory=lambda: {"lte": 100})
    path: Optional[dict[str, int]] = None
    row: Optional[dict[str, int]] = None


@dataclass
class Providers:
    """
    Represents the LGI-CDSR provider. The only provider supported by cbers4asat (by choice).
    """

    method: str = field(default="POST")
    name: str = field(default="LGI-CDSR")
    collections: list[Collection] = None
    query: Query = field(default_factory=Query)


@dataclass
class STACRequestBody(SerializationCapabilities):
    """
    Represents the STAC API request. This object will be send inside POST request body to INPE STAC API.
    """

    bbox: list[float] = field(default_factory=lambda: [-81, -37, -30, 11])
    datetime: str = None
    fromCatalog: str = field(default="yes")
    limit: int = 100
    providers: List[Providers] = field(default_factory=list)


@dataclass
class STACItemRequestBody:
    """
    Represents the STAC API request to get single/multiple item(s) full metadata.
    """

    ids: list[str] = None
    collection: Union[str, Collections] = None

    @property
    def empty(self):
        return self.collection is None and self.ids is None



# -*- coding: utf-8 -*-
# Standard Libraries
from dataclasses import dataclass
from typing import Union, Optional, TypeVar

# Local Modules
from .search import SearchItem
from .utils.dataclass import SerializationCapabilities, ignore_extras


@ignore_extras
@dataclass
class Geometry:
    """
    Class that represents the Item geometry interface.
    """

    type: str
    coordinates: list[list[list[float]]]


@ignore_extras
@dataclass
class Properties(SerializationCapabilities):
    """
    Class that represents the Item properties.
    """

    datetime: str
    path: int
    row: int
    satellite: str
    sensor: str
    cloud_cover: Union[float, int]


@ignore_extras
@dataclass
class Asset:
    """
    Class that represents an Asset from Item.
    """

    href: str
    type: str


@ignore_extras
@dataclass
class Assets:
    """
    Class that represents the assets from Item.
    """

    thumbnail: Asset
    red: Optional[Asset] = None
    green: Optional[Asset] = None
    blue: Optional[Asset] = None
    nir: Optional[Asset] = None
    pan: Optional[Asset] = None

    def __post_init__(self):
        if isinstance(self.thumbnail, dict):
            self.thumbnail = Asset(**self.thumbnail)
        if isinstance(self.red, dict):
            self.red = Asset(**self.red)
        if isinstance(self.green, dict):
            self.green = Asset(**self.green)
        if isinstance(self.blue, dict):
            self.blue = Asset(**self.blue)
        if isinstance(self.nir, dict):
            self.nir = Asset(**self.nir)
        if isinstance(self.pan, dict):
            self.pan = Asset(**self.pan)


Item = TypeVar("Item")


@ignore_extras
@dataclass
class Item(SerializationCapabilities):
    """
    Class to parse items from INPE STAC Catalog

    This represents one image from CBERS-4/CBERS-4A/AMAZONIA-1.
    """

    type: str
    id: str
    collection: str
    geometry: Geometry
    bbox: list[float]
    properties: Properties
    assets: Assets

    def __post_init__(self):
        if isinstance(self.geometry, dict):
            self.geometry = Geometry(**self.geometry)
        if isinstance(self.properties, dict):
            self.properties = Properties(**self.properties)
        if isinstance(self.assets, dict):
            self.assets = Assets(**self.assets)

    @staticmethod
    def from_search(_id: str, collection: str) -> Item | Exception:
        """
        Create an Item by making a search by ID inside a collection.

        Args:
            _id: Item ID
            collection: Collection to search into.
        Return:
            Item object.
        Raise:
            ``Exception`` if item not found.
        """
        search = SearchItem()
        search.ids(
            list([_id]),
            collection=collection,
        )

        features = search().get("features")

        if not len(features):
            raise Exception("Item not found.")

        return Item(**features[0])

    def get_assets(self) -> None:
        """
        Get assets/bands of the object.
        """
        self.assets = Item.from_search(self.id, self.collection).assets

    def has_band(self, band: str) -> bool:
        """
        Check if Item has a band search by its name.

        Args:
            band: Band name. Ex.: red
        Return:
            True if item has band.
        """
        return getattr(self.assets, band, None) is not None

    def band_url(self, band: str) -> str | Exception:
        """
        Get asset/band URL.

        Args:
            band: Band name. Ex.: red
        Return:
            URL
        Raise:
            ``Exception`` if Item does not have band.
        """
        if not self.has_band(band):
            raise Exception(f"Band {band} does not exist in this item!")
        return getattr(self.assets, band).href



# -*- coding: utf-8 -*-
# Standard Libraries
from datetime import date
from typing import Union

# PyPi Packages
from requests import Session, HTTPError

# Local Modules
from .collections import Collections
from .request import (
    STACRequestBody,
    Providers,
    Collection,
    STACItemRequestBody,
)


class SearchItem:
    """
    Simple class to search Item inside INPE STAC Catalog
    """

    # INPE STAC search item in collection
    BASE_URL_SEARCH_ITEM: str = "https://www.dgi.inpe.br/lgi-stac/collections"

    def __init__(self) -> None:
        self.search_item_body: STACItemRequestBody = STACItemRequestBody()

    def __call__(self) -> dict | Exception:
        """
        Make request using the search parameters.

        Return:
            GeoJson-like dictionary.
        Raise:
            ``Exception`` if any http error.
        """
        features = list()
        with Session() as session:
            for id_ in self.search_item_body.ids:
                try:
                    response = session.get(
                        f"{self.BASE_URL_SEARCH_ITEM}/{self.search_item_body.collection}/items/{id_}"
                    )
                    response.raise_for_status()
                    feature = response.json()
                    if feature.get("type") == "Feature":
                        features.append(feature)
                except HTTPError as err:
                    raise Exception(
                        f"{response.status_code} - ERROR searching {id_}. Reason: {response.reason}. Exception: {err}"
                    )

        return {"type": "FeatureCollection", "features": features}

    def ids(
        self, ids: list[str], collection: Union[str, Collections]
    ) -> None | Exception:
        """
        Id(s) to search inside a collection.

        Args:
            ids: Item id String or list of item id strings.
            collection: Collection name to search into as string or Collections Enum.
        Return:
            Void
        Raise:
            ``Exception`` if id(s) or collection is empty.
        """
        if not len(ids):
            raise Exception("Ids to search list cannot be empty.")
        elif not collection:
            raise Exception("Collection cannot be empty.")

        self.search_item_body.ids = ids
        self.search_item_body.collection = collection


class Search:
    """
    Simple class to search INPE STAC Catalog
    """

    # INPE STAC Catalog
    BASE_URL_SEARCH: str = "https://www.dgi.inpe.br/stac-compose/stac/search/"

    def __init__(self) -> None:
        self.stac_request_body = STACRequestBody()
        self.providers_body = Providers()

    def __call__(self) -> dict | Exception:
        """
        Make request using the search parameters.

        Return:
            GeoJson-like dictionary.
        Raise:
            ``Exception`` if any http error.
        """
        self.stac_request_body.providers.append(self.providers_body)
        with Session() as session:
            try:
                response = session.post(
                    self.BASE_URL_SEARCH,
                    json=self.stac_request_body.asdict(exclude_none=True),
                )
                response.raise_for_status()
                # Response Root Keys are the providers, like: "LGI-CDSR', "DATA-INPE"...
                # Get the only provider that will be supported by cbers4asat lib.
                collections = response.json().get("LGI-CDSR", None)
                # Second level of keys are the collections, like "AMAZONIA1_WFI_L2_DN".
                # Every collection will be grouped inside this variable bellow.
                feature_collection = {"type": "FeatureCollection", "features": []}

                if not collections:
                    return feature_collection

                # For every collection...
                for name, content in collections.items():
                    if not isinstance(content, dict):
                        continue

                    if not content.get("features", None):
                        continue

                    # Append all collection features in one
                    feature_collection["features"].extend(content["features"])
                return feature_collection
            except HTTPError as err:
                raise Exception(
                    f"{response.status_code} - ERROR in query. Reason: {response.reason}. Exception: {err}"
                )

    def bbox(self, bbox: list[float]) -> None | Exception:
        """
        Only products that have a geometry/footprint that intersects the bounding box are selected.

        Args:
            bbox: The bounding box provided as a list of four floats, minimum longitude, minimum latitude, maximum longitude and maximum latitude.
        Return:
            Void
        Raise:
            ``Exception`` if bbox list is empty or bbox coordinates is not float numbers or bbox does not have 4 coordinates.
        """
        if not len(bbox):
            raise Exception("Bounding box cannot be empty.")
        elif not all([isinstance(coord, float) for coord in bbox]):
            raise Exception("Bounding box coordinates must be float numbers.")
        elif len(bbox) != 4:
            raise Exception(
                "Bounding box must have four float numbers representing the coordinates."
            )

        self.stac_request_body.bbox = bbox

    def date_interval(self, start: date, end: date) -> None | Exception:
        """
        Search for scenes that was taken in this interval of dates.

        Args:
            start: Initial date. The older part of interval.
            end: End date. The most recent part of interval.
        Return:
            Void
        Raise:
            ``Exception`` if the start date is more recent than end date.
        """
        if start > end:
            raise Exception("Initial date must be older than the end date.")

        self.stac_request_body.datetime = (
            f"{start.isoformat()}T00:00:00/{end.isoformat()}T23:59:00"
        )

    def collections(
        self, collections: Union[list[str], list[Collections]]
    ) -> None | Exception:
        """
        Will search products inside these collections.

        Args:
            collections: Collections name list to search into as string or Collections Enum.
        Return:
            Void
        Raise:
            ``Exception`` if collections list is empty.
        """
        if not len(collections):
            raise Exception("Collections cannot be empty.")

        self.providers_body.collections = [
            Collection(name=collection) for collection in collections
        ]

    def limit(self, limit: int) -> None | Exception:
        """
        How much products will return from query.

        Args:
            limit: limit value of products that will return from query
        Return:
            Void
        Raise:
            ``Exception`` if limit value is less or equal than zero.
        """
        if limit <= 0:
            raise Exception("Limit value must be greater than 0.")

        self.stac_request_body.limit = limit

    def path_row(self, path: int, row: int) -> None | Exception:
        """
        Get product by specifying the cell from grid.
        Grid can be found: http://www.obt.inpe.br/OBT/assuntos/catalogo-cbers-amz-1

        Args:
            path: Path number from grid
            row: Row number from grid
        Return:
            Void
        Raise:
            ``Exception`` if path or row is empty.
        """
        if not path or not row:
            raise Exception("Path and/or row cannot be empty.")

        self.providers_body.query.path = dict(eq=path)
        self.providers_body.query.row = dict(eq=row)

    def cloud_cover(self, cloud_cover: int) -> None | Exception:
        """
        Maximum cloud coverage on scene.

        Args:
             cloud_cover: Maximum cloud cover percentage between 0 and 100 range.
        Return:
            Void
        Raise:
            ``Exception`` if cloud cover is less than zero and greater than 100.
        """
        if cloud_cover < 0 or cloud_cover > 100:
            raise Exception("Cloud cover must be between 0 and 100.")

        self.providers_body.query.cloud_cover.update(lte=cloud_cover)



# -*- coding: utf-8 -*-
# Standard Libraries
from enum import StrEnum


class Collections(StrEnum):
    AMAZONIA1_WFI_L2_DN = "AMAZONIA1_WFI_L2_DN"
    AMAZONIA1_WFI_L4_DN = "AMAZONIA1_WFI_L4_DN"
    CBERS4A_MUX_L2_DN = "CBERS4A_MUX_L2_DN"
    CBERS4A_MUX_L4_DN = "CBERS4A_MUX_L4_DN"
    CBERS4A_WFI_L2_DN = "CBERS4A_WFI_L2_DN"
    CBERS4A_WFI_L4_DN = "CBERS4A_WFI_L4_DN"
    CBERS4A_WPM_L2_DN = "CBERS4A_WPM_L2_DN"
    CBERS4A_WPM_L4_DN = "CBERS4A_WPM_L4_DN"
    CBERS4_AWFI_L2_DN = "CBERS4_AWFI_L2_DN"
    CBERS4_AWFI_L4_DN = "CBERS4_AWFI_L4_DN"
    CBERS4_MUX_L2_DN = "CBERS4_MUX_L2_DN"
    CBERS4_MUX_L4_DN = "CBERS4_MUX_L4_DN"
    CBERS4_PAN10M_L2_DN = "CBERS4_PAN10M_L2_DN"
    CBERS4_PAN10M_L4_DN = "CBERS4_PAN10M_L4_DN"
    CBERS4_PAN5M_L2_DN = "CBERS4_PAN5M_L2_DN"
    CBERS4_PAN5M_L4_DN = "CBERS4_PAN5M_L4_DN"

    def __repr__(self) -> str:
        return str.__repr__(self.value)



# -*- coding: utf-8 -*-
from dataclasses import dataclass, asdict as dataclass_as_dict
from typing import Any


def ignore_extras(cls):
    """
    Decorator to Ignore extra arguments inside dataclass constructor.

    This will prevent this exception to raise. Example:
    - TypeError: __init__() got an unexpected keyword argument 'stac_version'

    """
    __original__init__ = cls.__init__

    def filter_extras(self, **kwargs) -> None:
        """
        docstring
        """
        __original__init__(
            self, **{k: v for k, v in kwargs.items() if k in self.__annotations__}
        )

    cls.__init__ = filter_extras
    return cls


@dataclass
class SerializationCapabilities:
    """
    Serialization methods to dataclass.
    """

    def __del_none(self, value: dict) -> list[dict] | dict[Any, dict] | dict:
        """
        Delete keys with the value None in a dictionary, recursively.
        """
        if isinstance(value, list):
            return [self.__del_none(x) for x in value if x is not None]
        elif isinstance(value, dict):
            return {
                key: self.__del_none(val)
                for key, val in value.items()
                if val is not None
            }
        else:
            return value

    def asdict(self, exclude_none: bool = False) -> dict:
        """
        Dataclass to dictionary serialization. With exclude none values option.
        """
        if exclude_none:
            return self.__del_none(dataclass_as_dict(self).copy())
        return dataclass_as_dict(self).copy()



# -*- coding: utf-8 -*-
# Standard Libraries
from dataclasses import dataclass
from typing import Iterable

# Local Modules
from .item import Item
from .utils.dataclass import ignore_extras, SerializationCapabilities


@ignore_extras
@dataclass
class ItemCollection(SerializationCapabilities):
    """
    Class to parse response from INPE STAC Catalog.

    In this context, STAC Catalog uses the GeoJSON structure and has predefined properties keys.

    So, the ItemCollection class maps the outer container: The Collection and his items.
    """

    features: list[Item]
    type: str = "FeatureCollection"

    def __post_init__(self):
        if len(self.features):
            theres_a_dict = any(map(lambda f: isinstance(f, dict), self.features))

            if theres_a_dict:  # If there's a dict, do conversion to Item object.
                items = list()
                for feature in self.features:
                    if isinstance(feature, dict):
                        items.append(Item(**feature))
                    elif isinstance(feature, Item):
                        items.append(feature)
                    else:
                        raise TypeError(f"{feature} is an invalid feature type!")

                self.features = items.copy()

    def __iter__(self) -> Iterable[Item]:
        """
        Iterate the items inside this collection.
        """
        for feature in self.features:
            yield feature

    def get_features_assets(self) -> None:
        """
        STAC API return the items without assets, so, when needed, call this
        function to load all the assets inside every item.
        """
        for feature in self.features:
            feature.get_assets()



# -*- coding: utf-8 -*-
# Standard Libraries
from os import makedirs
from os.path import join, basename, exists

# PyPi Packages
from requests import Session, HTTPError
from requests.adapters import HTTPAdapter
from urllib3.util import Retry


class Download:
    """
    Class to download assets from INPE STAC Catalog.
    """

    def __init__(self):
        retries = Retry(
            total=3,
            connect=3,
            read=3,
            status=3,
            other=3,
            backoff_factor=1,
            status_forcelist=[500, 501, 502, 503, 504],
            allowed_methods={"GET"},
        )
        self.session = Session()
        self.session.mount("http://", HTTPAdapter(max_retries=retries))

    def download(self, url: str, credential: str, outdir: str) -> None | Exception:
        """
        Download the asset.

        Args:
            url: URL pointing to asset/band .TIFF
            credential: e-mail used in the explorer inpe platform.
            outdir: Output directory
        Raise:
            ``Exception`` if any http error occurs.
        """
        if not exists(outdir):
            makedirs(outdir, exist_ok=True)

        geotiff = basename(url)

        outfile = join(outdir, geotiff)

        with self.session as session:
            try:
                response = session.get(
                    url,
                    params={"email": credential},
                    stream=True,
                    allow_redirects=True,
                )
                response.raise_for_status()
                with open(outfile, "wb") as f:
                    for chunk in response.iter_content(chunk_size=4096):
                        if chunk:
                            f.write(chunk)
            except HTTPError as err:
                raise Exception(
                    f"{response.status_code} - ERROR in {url}. Reason: {response.reason}. Exception: {err}"
                )
